from rich.console import Console
from rich.table import Table

from .argparser import parse
from .models.openai import OpenAIClient
from .library import (
    save_to_library
)
from .utils import (
    execute_unix_command, 
    copy_command_to_clipboard
)
from .prompts import OPENAI_UNIX_COMMAND_PROMPT


ACTIONS = {
    "execute": "e",
    "copy": "c",
    "save": "s",
    "revise": "r",
    "abort": "a"
}

console = Console()


def cli(OPEN_API_KEY):
    args = parse()

    # -i, --input
    input: str = args.input.strip()
    if input:
        handle_input(input, OPEN_API_KEY)


def handle_input(nl_input: str, OPEN_API_KEY: str):
    openai_client = OpenAIClient(OPEN_API_KEY)

    console.print("=> Searching... 👨‍💻")
    unix_command = openai_client.fetch_unix_command(
        user_input=nl_input,
        system_prompt=OPENAI_UNIX_COMMAND_PROMPT,
        model="gpt-3.5-turbo",
    )
    unix_command = unix_command.strip()

    if unix_command.lower() == "unknown":
        console.print("=> Command not found. 🤷‍♂️")

        table = Table()
        table.add_column("Action", style="bold")
        table.add_column("Key", style="bold")
        table.add_column("Description", style="bold")

        actions = [
            {"Name": "Revise", "Key": "r", "Description": "Revise the input"},
            {"Name": "Abort", "Key": "a", "Description": "Abort the program"},
        ]

        for action in actions:
            table.add_row(
                action["Name"],
                action["Key"],
                action["Description"],
            )

        console.print(table)

        handle_unknown_actions(nl_input=nl_input)

    else:
        console.print(f"=> Command found! 🙆‍♂️ [code] {unix_command} [/code]")

        table = Table()
        table.add_column("Action", style="bold")
        table.add_column("Key", style="bold")
        table.add_column("Combineable", style="bold")
        table.add_column("Description", style="bold")

        actions = [
            {"Name": "Execute", "Key": "e", "Combineable": "Yes", "Description": "Execute the command"},
            {"Name": "Copy", "Key": "c", "Combineable": "Yes", "Description": "Copy the command"},
            {"Name": "Save", "Key": "s", "Combineable": "Yes", "Description": "Save the command"},
            {"Name": "Revise", "Key": "r", "Combineable": "No", "Description": "Revise the input"},
            {"Name": "Abort", "Key": "a", "Combineable": "No", "Description": "Abort the program"},
        ]

        for action in actions:
            table.add_row(
                action["Name"],
                action["Key"],
                action["Combineable"],
                action["Description"],
            )

        console.print(table)

        handle_known_actions(unix_command=unix_command, nl_input=nl_input)


def handle_unknown_actions(nl_input: str):
    action_input = input("=> Choose an action key: ").strip().lower()
    
    if not any(act in action_input for act in ACTIONS.values()):
        console.print("=> Enter a valid action. 🤦‍♂️")
        handle_unknown_actions(nl_input)
    
    if ACTIONS["abort"] in action_input:
        handle_abort_action()
        return
    if ACTIONS["revise"] in action_input:
        handle_revise_action(nl_input)
        return


def handle_known_actions(unix_command: str, nl_input: str):
    action_input = input("=> Choose action key(s): ").strip().lower()

    if not any(act in action_input for act in ACTIONS.values()):
        console.print("=> Enter a valid action. 🤦‍♂️")
        handle_known_actions(unix_command, nl_input)

    if ACTIONS["abort"] in action_input:
        handle_abort_action()
        return
    if ACTIONS["revise"] in action_input:
        handle_revise_action(nl_input)
        return
    
    if ACTIONS["copy"] in action_input:
        handle_copy_action(unix_command)
    if ACTIONS["save"] in action_input:
        handle_save_action(unix_command, nl_input)
    if ACTIONS["execute"] in action_input:
        handle_execute_action(unix_command)



def handle_execute_action(unix_command: str):
    console.print("=> Executing... 👨‍💻")
    response = execute_unix_command(unix_command)

    if "success" in response:
        console.print(f"=> {response['success']}")
    else:
        console.print(f"=> {response['error']}")



def handle_copy_action(unix_command: str):
    response = copy_command_to_clipboard(unix_command)

    if "success" in response:
       console.print(f"=> {response['success']}")
    else:
        console.print(f"=> {response['error']}")


def handle_save_action(unix_command: str, nl_input: str):
    save_to_library(nl_input, unix_command)


def handle_revise_action(nl_input: str):
    pass


def handle_abort_action():
    console.print("=> Aborting... 🙍‍♂️")
    exit(1)


# ======================================================
# UnixGPT Default Values Collection
# ======================================================
#
# Default values used throughout the entire library.

# Library version
VERSION: str = "0.1.4"

# Default model for the library
DEFAULT_MODEL: str = "gpt-3.5-turbo"

# Current supported models
SUPPORTED_MODELS: list[str] = ["gpt-3.5-turbo"]



import os

from .cli import cli


OPENAI_API_KEY: str = os.environ.get("OPENAI_API_KEY", None)
if not OPENAI_API_KEY:
    raise ValueError("OPENAI_API_KEY environment variable is required.")


def main():
    cli(OPENAI_API_KEY)


if __name__ == "__main__":
    try:
        main()
    except (EOFError, KeyboardInterrupt):
        exit()


import argparse

from .constants import VERSION


def parse():
    parser = argparse.ArgumentParser(
        prog="unixgpt",
        description="Bridges the gap between natural language and UNIX commands"
    )

    # collect input (required)
    parser.add_argument(
        "-i", "--input",
        type=str,
        help="Natural language input to be converted to unix command. Enclose in quotation marks."
    )

    # show saved commands in unixgpt local library
    parser.add_argument(
        "-l", "--list",
        action="store_true",
        help="List saved commands in UnixGPT local library"
    )

    # save command to unixgpt local library
    parser.add_argument(
        "-s", "--save",
        type=str,
        help="Store a command in UnixGPT local library for easy access"
    )

    # fetch a command from unixgpt local library with natural language or by ID
    parser.add_argument(
        "-f", "--fetch",
        type=str,
        help="Fetch a command from UnixGPT local library with natural language or by ID"
    )

    # delete a saved command from unixgpt local library by ID
    parser.add_argument(
        "-d", "--delete",
        type=str,
        help="Delete a saved command from UnixGPT local library"
    )

    # unixgpt version
    parser.add_argument(
        "--version",
        action="version",
        version=VERSION,
        help="Print version information and quit"
    )
    
    args = parser.parse_args()
    return args



import subprocess
import pyperclip

def execute_unix_command(unix_command: str) -> dict:
    try:
        subprocess.run(unix_command, shell=True, check=True)
        return {"success": f"Command '{unix_command}' ran successfully."}
    except subprocess.SubprocessError as e:
        return {"error": f"Command '{unix_command}' failed to run."}
    except Exception as e:
        return {"error": f"An unexpected error occurred while executing."}


def copy_command_to_clipboard(unix_command: str) -> dict:
    try:
        pyperclip.copy(unix_command)
        return {"success": f"Command '{unix_command}' copied to clipboard successfully."}
    except pyperclip.PyperclipException:
        return {"error": f"Command '{unix_command}' was not able to be copied."} 
    except Exception as e:
        return {"error": f"An unexpected error occurred while copying."}



import openai

from ..constants import DEFAULT_MODEL
from ..prompts import DEFAULT_PROMPT

class OpenAIClient:
    def __init__(self, OPENAI_API_KEY):
        openai.api_key = OPENAI_API_KEY
        
        # model prices as of 10/31/23
        self.prices = {
            "gpt-3.5-turbo-4k": [0.0015, 0.002],
            "gpt-3.5-turbo-16k": [0.003, 0.004],
            "ft-gpt-3.5-turbo": [0.012, 0.016],
            "ft-babbage-002": [0.0016, 0.0016],
            "ft-davinci-002": [0.012, 0.012],
            "gpt-4-8k": [0.03, 0.06],
            "gpt-4-32k": [0.06, 0.12],
        }

    def fetch_unix_command(
            self, 
            user_input: str = "", 
            system_prompt: str = DEFAULT_PROMPT,
            model: str = DEFAULT_MODEL,
            temperature: float = 0,
            max_tokens: int = 250,
        ) -> str:
        """
        Converts natural language user input to a corresponding UNIX command.
        """
        
        unix_command: str = self._fetch_from_gpt(
            user_input=user_input,
            system_prompt=system_prompt,
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
        )
        
        return unix_command.strip()

    def _fetch_from_gpt(
            self,
            user_input,
            system_prompt,
            model, 
            temperature,
            max_tokens,
        ) -> str:
        try:
            completion = openai.ChatCompletion.create(
                model=model,
                temperature=temperature,
                max_tokens=max_tokens,
                messages=[
                    {"role": "system", "content": system_prompt},
                    {"role": "user", "content": user_input}
                ]
            )
        except Exception as e:
            print(e)
            
        try:
            result = completion.choices[0].message.content
        except KeyError as e:
            print(e)

        return result
    
    def get_total_cost_of_tokens_used(
            self, 
            model,
            user_query: str, 
            system_prompt: str,
            output: str,
            encoding_name: str = "cl100k_base"
        ) -> float:
        """
        Get total cost of using an OpenAI model for a given request.
        """

        try:
            price_per_1k_tokens: list = self.prices[model]
        except KeyError as e:
            print("Model does not exist.")
            return

        price_per_token_input = price_per_1k_tokens[0] / 1000
        price_per_token_output = price_per_1k_tokens[1] / 1000

        input_content: str = user_query + "\n" + system_prompt
        num_tokens_input = self.get_num_of_tokens(
            content=input_content,
            encoding_name=encoding_name
        )
        num_tokens_output = self.get_num_of_tokens(
            content=output,
            encoding_name=encoding_name
        )

        estimated_cost_input = price_per_token_input * num_tokens_input
        estimated_cost_output = price_per_token_output * num_tokens_output

        total_cost = estimated_cost_input + estimated_cost_output
        return round(total_cost, 3)

    def get_num_of_tokens(
            self,
            content: str,
            encoding_name: str = "cl100k_base",
        ):
        """
        Determine the number of tokens in a given string.
        """

        import tiktoken

        encoding = tiktoken.get_encoding(encoding_name=encoding_name)
        num_tokens = len(encoding.encode(content))

        return num_tokens


DEFAULT_PROMPT = """
"""

OPENAI_UNIX_COMMAND_PROMPT = """
Imagine you are a UNIX command expert, and your task is to translate user input into accurate UNIX commands. 
Your goal is to provide the complete UNIX command(s) that best address the user's request. 
Be aware that a user may issue multiple commands in a single input, so use the '&&' UNIX command to separate them as necessary. 
Pay special attention to unique names provided by the user and create any necessary directories and files (e.g. rest/app.py).
Your response should include only the UNIX command(s) that fully meet the user's query without any additional information.
If you are unable to answer the question or it is not relevant to the above instructions, simply state 'unknown'
"""

OPENAI_FETCH_FROM_LIB_PROMPT = """
"""


import os

from setuptools import setup, find_packages

from unixgpt.constants import VERSION

# dependencies
REQUIREMENTS = [
    "openai==0.28.1",
    "pyperclip==1.8.2",
    "rich==13.6.0",
]

setup(
    name="unixgpt",
    version=VERSION,
    packages=find_packages(),
    install_requires=REQUIREMENTS,
    entry_points={
        "console_scripts": [
            "unixgpt = unixgpt.__main__:main"
        ]
    },
    description="unixgpt is a powerful bridge between natural language and UNIX commands",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
)
